#include <string>
#include <boost/algorithm/string.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include "cookie_manager.h"
#include "cookie.h"
#include "utils.h"
#include "system.h"

namespace simple_http
{
	void cookie_manager::add_cookie(const std::string& name, const cookie& value)
	{
		std::pair<std::multimap<std::string, cookie>::const_iterator, std::multimap<std::string, cookie>::const_iterator> ret = 
			cookies_.equal_range(name);

		bool duplicate = false;

		for(std::multimap<std::string, cookie>::const_iterator i = ret.first; i != ret.second; ++i)
		{
			if(boost::iequals((*i).second.domain, value.domain) &&
				boost::equals((*i).second.path, value.path))
			{
				duplicate = true;
				break;
			}
		}

		if(!duplicate)
			cookies_.insert(std::make_pair(name, value));
	}

	void cookie_manager::remove_cookie(const std::string& name, const std::string& domain, const std::string& path)
	{
		std::string domain_compare(domain);
		if(!utils::is_ip_address(domain_compare) && domain_compare[0] != '.')
			domain_compare = "." + domain_compare;

		std::multimap<std::string, cookie>::const_iterator i = cookies_.begin();
		while(i != cookies_.end())
		{
			if(boost::iequals((*i).first, name) &&
				(domain.empty() || boost::iequals((*i).second.domain, domain_compare)) &&
				(path.empty() || boost::equals((*i).second.path, path)))
				i = cookies_.erase(i);
			else
				++i;
		}
	}

	void cookie_manager::remove_expired_cookies()
	{
		std::multimap<std::string, cookie>::const_iterator i = cookies_.begin();
		while(i != cookies_.end())
		{
			if(!(*i).second.session &&
				boost::posix_time::ptime_from_tm(system::get_utc_time()) >=
				boost::posix_time::ptime_from_tm((*i).second.expires))
				i = cookies_.erase(i);
			else
				++i;
		}
	}

	void cookie_manager::remove_session_cookies()
	{
		std::multimap<std::string, cookie>::const_iterator i = cookies_.begin();
		while(i != cookies_.end())
		{
			if((*i).second.session)
				i = cookies_.erase(i);
			else
				++i;
		}
	}

	std::size_t cookie_manager::total_count() const
	{
		return cookies_.size();
	}

	void cookie_manager::add_replace_cookie(const std::string& name, const cookie& value)
	{
		remove_cookie(name, value.domain, value.path);
		add_cookie(name, value);
	}

	void cookie_manager::get_cookies(std::map<std::string, cookie>& cookies, const std::string& domain, const std::string& path, connection_type connection) const
	{
		bool ip = utils::is_ip_address(domain);

		for(std::multimap<std::string, cookie>::const_iterator i = cookies_.begin(); i != cookies_.end(); ++i)
		{
			if(!boost::istarts_with(path, (*i).second.path))
				continue;

			if((ip || utils::is_ip_address((*i).second.domain)))
			{
				if(domain != (*i).second.domain)
					continue;
			}
			else if(!boost::iends_with((*i).second.domain, "." + domain))
				continue;

			if(connection == connection_http && (*i).second.secure)
				continue;

			std::map<std::string, cookie>::iterator it = cookies.find((*i).first);
			if(it == cookies.end())
			{
				cookies.insert(*i);
			}
			else
			{
				if((*it).second.domain.length() <= (*i).second.domain.length()
					||
					(*it).second.path.length() <= (*i).second.path.length())
				{
					cookies.erase(it);
					cookies.insert(*i);
				}
			}
		}
	}

	std::string cookie_manager::get_cookies(const std::string& domain, const std::string& path, connection_type connection) const
	{
		std::string ret;
		std::map<std::string, cookie> cookies;
		get_cookies(cookies, domain, path, connection);
		for(std::map<std::string, cookie>::const_iterator it = cookies.begin(); it != cookies.end(); ++it)
			ret.append((*it).first).append("=").append((*it).second.value).append("; ");

		return ret;
	}

	void cookie_manager::clear_cookies()
	{
		cookies_.clear();
	}
}
