#include <string>
#include <map>
#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include "simple_http_defs.h"
#include "functions.h"
#include "static_data.h"
#include "cookie.h"
#include "cookie_manager.h"
#include "utils.h"
#include "system.h"

namespace simple_http
{
	void get_headers(const std::string& http_answer, std::map<std::string, std::string>& headers)
	{
		std::string parse = get_headers(http_answer);
		headers.clear();

		std::string::size_type pos = 0, oldpos = 0;
		while((pos = parse.find("\r\n", oldpos + 1)) != std::string::npos)
		{
			boost::smatch what;
			std::string possible_header(http_answer.c_str() + oldpos, pos - oldpos);
			boost::trim(possible_header);

			try
			{
				if(boost::regex_match(possible_header, what, header_regex))
				{
					if(!boost::iequals(what[1].str(), "set-cookie"))
						headers.insert(std::make_pair(what[1].str(), what[2].str()));
				}
			}
			catch(const boost::regex_error&)
			{
			}

			oldpos = pos;
		}
	}

	std::string get_headers(const std::string& http_answer)
	{
		std::string ret;
		std::string::size_type pos;
		if((pos = http_answer.find("\r\n\r\n")) != std::string::npos)
			ret = http_answer.substr(0, pos);

		return ret;
	}

	std::string get_body(const std::string& http_answer)
	{
		std::string ret;
		std::string::size_type pos;
		if((pos = http_answer.find("\r\n\r\n")) != std::string::npos)
			ret = http_answer.substr(pos + 4);

		return ret;
	}

	void get_cookies(const std::string& http_headers, cookie_manager& manager, const std::string& domain, const std::string& path, connection_type connection)
	{
		//Set-Cookie: lu=Rg3vHJZnehYLjVg7qi3bZjzg; Expires=Tue, 15-Jan-2013 21:47:38 GMT; Path=/; Domain=.foo.com; HttpOnly
		boost::match_results<std::string::const_iterator> what;
		boost::match_flag_type flags = boost::match_default;
		std::string::const_iterator start = http_headers.begin();

		try
		{
			while(boost::regex_search(start, http_headers.end(), what, cookie_regex, flags)) 
			{
				cookie cook;
				cook.value = what[3].str();

				bool valid = true,
					expired = false;

				{
					const std::string parameters = boost::trim_copy(what[4].str()) + ";";
					boost::match_results<std::string::const_iterator> params;
					boost::match_flag_type param_flags = boost::match_default;
					std::string::const_iterator param_start = parameters.begin();

					try
					{
						while(boost::regex_search(param_start, parameters.end(), params, cookie_param_regex, param_flags)) 
						{
							if(param_start != params[0].first)
							{
								valid = false;
								break;
							}

							if(boost::iequals(params[1].str(), "Expires"))
							{
								cook.expires = utils::cookie_date_to_tm(params[3].str());
								if(!cook.expires.tm_year)
								{
									valid = false;
									break;
								}

								if(boost::posix_time::ptime_from_tm(cook.expires) <= boost::posix_time::ptime_from_tm(system::get_utc_time()))
									expired = true;
							}
							else if(boost::iequals(params[1].str(), "Path"))
							{
								cook.path = params[3].str();
							}
							else if(boost::iequals(params[1].str(), "Domain"))
							{
								cook.domain = params[3].str();
								try
								{
									//Matches IP addresses also
									if(!boost::regex_match(cook.domain, domain_regex))
									{
										valid = false;
										break;
									}
								}
								catch(const boost::regex_error&)
								{
									valid = false;
									break;
								}
							}
							else if(boost::iequals(params[1].str(), "Max-Age"))
							{
								try
								{
									long max_age = boost::lexical_cast<long>(params[3].str());
									if(!max_age)
										expired = true;

									cook.expires = boost::posix_time::to_tm(
										(boost::posix_time::ptime_from_tm(system::get_utc_time()) + boost::posix_time::seconds(max_age)));
								}
								catch(const boost::bad_lexical_cast&)
								{
									valid = false;
									break;
								}
							}
							else if(boost::iequals(params[1].str(), "HttpOnly"))
							{
								cook.httponly = true;
							}
							else if(boost::iequals(params[1].str(), "Secure"))
							{
								if(connection != connection_http)
								{
									cook.secure = true;
								}
								else
								{
									valid = false;
									break;
								}
							}

							param_start = params[0].second;
							param_flags |= boost::match_prev_avail;
							param_flags |= boost::match_not_bob;
						}
					}
					catch(const boost::regex_error&)
					{
						valid = false;
					}

					if(param_start != parameters.end())
						valid = false;
				}


				while(valid)
				{
					if(cook.expires.tm_yday == 0)
						cook.session = true;

					if(cook.path.empty())
						cook.path = "/";

					if(!path.empty())
					{
						//* The value for the Path attribute is not a prefix of the request-URI.
						if(!boost::starts_with(path, cook.path))
							break;
					}

					if(!domain.empty())
					{
						if(utils::is_ip_address(cook.domain) || utils::is_ip_address(domain))
						{
							if(cook.domain.empty())
								cook.domain = domain;
							else if(cook.domain != domain)
								break;
						}
						else
						{
							if(cook.domain.empty())
								cook.domain = "." + domain;

							//* The value for the request-host does not domain-match the Domain attribute.
							if(!boost::iends_with("." + domain, cook.domain))
								break;

							{
								//  Let's do as browsers do:
								if(cook.domain[0] != '.')
									cook.domain = "." + cook.domain;

								//* The value for the Domain attribute contains no embedded dots or
								//  does not start with a dot.
								//  Also, check for localhost
								if(!boost::iequals(cook.domain, ".localhost"))
								{
									std::string::size_type pos = cook.domain.find_last_of('.');
									if(cook.domain[0] != '.' || pos == 0 || pos == cook.domain.length())
										break;
								}
							}

							if(domain.length() > cook.domain.length())
							{
								//* The request-host is a FQDN (not IP address) and has the form HD,
								//  where D is the value of the Domain attribute, and H is a string
								//  that contains one or more dots.

								boost::iterator_range<std::string::const_iterator> range = 
									boost::find_head(domain, 0 - cook.domain.length());

								if(std::string(range.begin(), range.end()).find('.') != std::string::npos)
									break;
							}
						}
					}

					if(expired)
						manager.remove_cookie(what[2].str(), cook.domain, cook.path);
					else
						manager.add_replace_cookie(what[2].str(), cook);

					break;
				}

				start = what[0].second;

				flags |= boost::match_prev_avail;
				flags |= boost::match_not_bob;
			}
		}
		catch(const boost::regex_error&)
		{
		}
	}

	std::string get_cookies(const std::string& http_headers, const std::string& domain, const std::string& path, connection_type connection)
	{
		cookie_manager cookies;
		get_cookies(http_headers, cookies, domain, path, connection);
		return cookies.get_cookies(domain, path);
	}

	unsigned long get_answer_code(const std::string& http_headers)
	{
		//HTTP/1.x 200
		if(http_headers.length() < 12)
			return 0;

		std::string::size_type pos = http_headers.find(' ');
		if(pos == std::string::npos)
			return 0;

		try
		{
			return boost::lexical_cast<unsigned long>(http_headers.substr(pos + 1, 3));
		}
		catch(const boost::bad_lexical_cast&)
		{
			return 0;
		}
	}
}
