#include <fstream>
#include <boost/scoped_array.hpp>
#include <boost/noncopyable.hpp>
#include "multipart_file.h"
#include "http_error.h"

namespace simple_http
{
	class multipart_file_impl : boost::noncopyable
	{
	public:
		multipart_file_impl(const std::string& name, const std::string& filename, const std::wstring& filename_w);
		boost::scoped_array<char> data_;
		unsigned long data_length_;
		unsigned long internal_ptr_;
		std::string filename_;
		std::wstring filename_w_;
		std::string name_;

		std::ifstream file_;
	};

	multipart_file_impl::multipart_file_impl(const std::string& name, const std::string& filename, const std::wstring& filename_w)
		:name_(name), filename_(filename), filename_w_(filename_w), data_length_(0), internal_ptr_(0)
	{}

	multipart_file::multipart_file(const std::string& name, const std::string& filepath)
		:impl(new multipart_file_impl(name, filepath, L""))
	{}

	multipart_file::multipart_file(const std::string& name, const std::wstring& filepath)
		:impl(new multipart_file_impl(name, "", filepath))
	{}

	void multipart_file::close()
	{
		reset_pointer();
	}

	unsigned long multipart_file::get_length() const
	{
		if(impl->data_.get())
		{
			return impl->data_length_;
		}
		else
		{
			std::ifstream file;
			if(impl->filename_.empty())
				file.open(impl->filename_w_, std::ios::in | std::ios::binary);
			else
				file.open(impl->filename_, std::ios::in | std::ios::binary);

			if(!file.is_open() || file.bad())
				throw http_error("Error opening HTTP multipart file", http_error::error_opening_multipart_file);

			file.seekg(0, std::ios::end);
			return static_cast<unsigned long>(file.tellg());
		}
	}


	multipart_file::multipart_file(const std::string& name, const char* binary_data, unsigned long length)
		:impl(0)
	{
		try
		{
			impl = new multipart_file_impl(name, "", L"");
			impl->data_length_ = length;
			impl->data_.reset(new char[length]);
			memcpy(impl->data_.get(), binary_data, length);
		}
		catch(const std::exception&)
		{
			delete impl;
			throw;
		}
	}

	void multipart_file::reset_pointer()
	{
		impl->internal_ptr_ = 0;
		impl->file_.clear();
		impl->file_.close();
	}

	unsigned long multipart_file::read_data(char* data, unsigned long length)
	{
		if(impl->data_.get())
		{
			unsigned long ret = std::min(impl->data_length_ - impl->internal_ptr_, length);
			memcpy(data, impl->data_.get() + impl->internal_ptr_, ret);
			impl->internal_ptr_ += ret;
			return ret;
		}
		else
		{
			if(!impl->file_.is_open() || impl->file_.bad())
			{
				if(impl->filename_.empty())
					impl->file_.open(impl->filename_w_, std::ios::in | std::ios::binary);
				else
					impl->file_.open(impl->filename_, std::ios::in | std::ios::binary);

				if(!impl->file_.is_open() || impl->file_.bad())
					throw http_error("Error opening HTTP multipart file", http_error::error_opening_multipart_file);

				impl->internal_ptr_ = 0;
			}

			impl->file_.read(data, length);
			return static_cast<unsigned long>(impl->file_.gcount());
		}
	}

	multipart_file::~multipart_file()
	{
		delete impl;
	}

	std::string multipart_file::get_name() const
	{
		return impl->name_;
	}
}
