/* Simple HTTP Class library By DX (http://kaimi.ru) */
/* Free to use or distribute */
#include <string>
#include <map>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include "simple_http.h"
#include "http_error.h"
#include "multipart_file.h"
#include "cookie.h"
#include "utils.h"
#include "cookie_manager.h"
#include "system.h"
#include "functions.h"
#include "simple_http_impl.h"
#include "http_endpoint.h"

namespace simple_http
{
	http_sync_connection::http_sync_connection()
		:impl(new simple_http_impl())
	{}

	http_sync_connection::http_sync_connection(const std::string& address, unsigned short port, bool ssl)
		:impl(new simple_http_impl(address, port, ssl))
	{}

	http_sync_connection::http_sync_connection(const http_endpoint& ep)
		:impl(new simple_http_impl(ep.address(), ep.port(), ep.ssl()))
	{}

	void http_sync_connection::set_connection_params(const http_endpoint& ep)
	{
#ifndef SSL_SUPPORT
		if(ep.ssl())
			throw http_error("SSL connections are not supported. Rebuild simple_http library with SSL_SUPPORT definition", http_error::ssl_not_supported);
#endif
		impl->address_ = ep.address();
		impl->port_ = ep.port();
		impl->ssl_ = ep.ssl();
	}

	http_sync_connection::~http_sync_connection()
	{
		delete impl;
	}

	std::string http_sync_connection::get_address() const
	{
		return impl->address_;
	}

	unsigned short http_sync_connection::get_port() const
	{
		return impl->port_;
	}

	std::string http_sync_connection::get_user_agent() const
	{
		return impl->user_agent_;
	}

	void http_sync_connection::set_user_agent(const std::string& user_agent)
	{
		impl->user_agent_ = user_agent;
	}

	bool http_sync_connection::is_connection_cache_enabled() const
	{
		return impl->connect_cache_enabled_;
	}

	void http_sync_connection::set_connenction_cache(bool enabled)
	{
		impl->connect_cache_enabled_ = enabled;
		if(!enabled)
			impl->connect_endpoint_cached_ = false;
	}

	void http_sync_connection::set_connection_params(const std::string& address, unsigned short port, bool ssl)
	{
		impl->address_ = address;
		impl->port_ = port;
		impl->connect_endpoint_cached_ = false;
		impl->ssl_ = ssl;

#ifndef SSL_SUPPORT
		if(ssl)
			throw http_error("SSL connections are not supported. Rebuild simple_http library with SSL_SUPPORT definition", http_error::ssl_not_supported);
#endif
	}

	std::string http_sync_connection::get(const std::string& uri, const std::string& parameters, const std::string& cookies, const std::string& referer)
	{
		return query("GET", uri, parameters, cookies, referer);
	}

	std::string http_sync_connection::post(const std::string& uri, const std::string& parameters, const std::string& cookies, const std::string& referer)
	{
		return query("POST", uri, parameters, cookies, referer);
	}

	std::string http_sync_connection::multipart_post(const std::string& uri, const std::string& multipart_file_parameter_name, multipart_file& file, const std::string& parameters, const std::string& cookies, const std::string& referer)
	{
		std::map<std::string, multipart_file*> files;
		files.insert(std::make_pair(multipart_file_parameter_name, &file));
		return multipart_post(uri, files, parameters, cookies, referer);
	}

	std::string http_sync_connection::multipart_post(const std::string& uri, const std::string& parameters, const std::string& cookies, const std::string& referer)
	{
		std::map<std::string, multipart_file*> files;
		return multipart_post(uri, files, parameters, cookies, referer);
	}

	std::string http_sync_connection::multipart_post(const std::string& uri, std::map<std::string, multipart_file*>& files, const std::string& parameters, const std::string& cookies, const std::string& referer)
	{
		if(impl->follow_redirects_)
		{
			impl->current_redirect_count_ = 0;
			impl->last_redirect_domain_.clear();
			impl->last_redirect_line_.clear();
			impl->last_redirect_port_ = 0;
			impl->last_redirect_uri_.clear();
		}

		return internal_query("POST", uri, parameters, cookies, referer, true, &files);
	}

	std::string http_sync_connection::internal_query(const std::string& method, const std::string& uri, const std::string& parameters, const std::string& cookies, const std::string& referer, bool multipart, std::map<std::string, multipart_file*>* files)
	{
		std::string ret;

		{
			boost::asio::streambuf buf;
			impl->query(method, buf, uri, parameters, cookies, referer, multipart, files, load_callback_);

			const char* header = boost::asio::buffer_cast<const char*>(buf.data());
			ret = std::string(header, buf.size());
			buf.consume(buf.size());
		}

		if(impl->cookie_mgr_)
			get_cookies(get_headers(ret), *impl->cookie_mgr_, impl->address_, utils::get_path_from_uri(uri), impl->ssl_ ? connection_secure : connection_http);

		if(impl->follow_redirects_ && impl->current_redirect_count_ < impl->max_redirect_count_)
		{
			std::map<std::string, std::string> headers;
			get_headers(ret, headers);
			std::map<std::string, std::string>::const_iterator it;
			if((it = headers.find("Location")) != headers.end())
			{
				std::string new_location = (*it).second;
				if(new_location.empty())
					return ret;

				impl->last_redirect_line_ = new_location;

				bool kill_referer = false;

				if(!boost::istarts_with(new_location, "http://")
#ifdef SSL_SUPPORT
					&& !boost::istarts_with(new_location, "https://")
#endif
					)
				{
					//ftp etc are not supported
					if(new_location.find("://") != std::string::npos)
						return ret;

					if('/' != new_location[0])
						new_location = utils::get_path_from_uri(uri) + new_location;
				}
				else
				{
					bool https_redirect = false;

					{
						unsigned long len = new_location.length();
						boost::ireplace_first(new_location, "http://", "");
						if(len == new_location.length())
						{
							https_redirect = true;
							boost::ireplace_first(new_location, "https://", "");
						}
					}

					std::string new_uri("/");

					std::string::size_type pos = new_location.find('/');
					if(pos != std::string::npos)
						new_uri = std::string(new_location.c_str() + pos);

					new_location = new_location.substr(0, pos);
					pos = new_location.find(':');

					std::string new_domain;
					unsigned short new_port;
					if(pos != std::string::npos)
					{
						try
						{
							new_domain = new_location.substr(0, pos);
							new_port = boost::lexical_cast<unsigned short>(std::string(new_location.c_str() + pos + 1));
						}
						catch(const boost::bad_lexical_cast&)
						{
							return ret;
						}
					}
					else
					{
						new_domain = new_location;
						new_port = 80;
					}

					if(!impl->allow_cross_domain_redirects_ && !boost::iequals(new_domain, impl->address_))
						return ret;
					
#ifdef SSL_SUPPORT
					if(is_ssl_connection() && !https_redirect)
						kill_referer = true;
#endif

					set_connection_params(new_domain, new_port, https_redirect);

					new_location = new_uri;
				}

				impl->current_redirect_count_++;
				impl->last_redirect_domain_ = impl->address_;
				impl->last_redirect_port_ = impl->port_;
				impl->last_redirect_uri_ = new_location;

				return internal_query("GET", new_location, "", "", kill_referer ? "" : referer);
			}
		}

		return ret;
	}

	std::string http_sync_connection::query(const std::string& method, const std::string& uri, const std::string& parameters, const std::string& cookies, const std::string& referer)
	{
		if(impl->follow_redirects_)
		{
			impl->current_redirect_count_ = 0;
			impl->last_redirect_domain_.clear();
			impl->last_redirect_line_.clear();
			impl->last_redirect_port_ = 0;
			impl->last_redirect_uri_.clear();
		}

		return internal_query(method, uri, parameters, cookies, referer);
	}

	std::string http_sync_connection::get_proxy_address() const
	{
		return impl->proxy_address_;
	}

	bool http_sync_connection::is_ssl_connection() const
	{
		return impl->ssl_;
	}

	std::string http_sync_connection::get_proxy_login() const
	{
		return impl->proxy_login_;
	}

	std::string http_sync_connection::get_proxy_password() const
	{
		return impl->proxy_password_;
	}

	unsigned short http_sync_connection::get_proxy_port() const
	{
		return impl->proxy_port_;
	}

	proxy_type http_sync_connection::get_proxy_type() const
	{
		return impl->proxy_type_;
	}

	void http_sync_connection::set_proxy(const std::string& address, unsigned short port, proxy_type ptype)
	{
		impl->proxy_address_ = address;
		impl->proxy_port_ = port;
		impl->proxy_type_ = ptype;
		impl->connect_endpoint_cached_ = false;
	}

	void http_sync_connection::set_proxy(const std::string& address, unsigned short port, proxy_type ptype, const std::string& login, const std::string& password)
	{
		impl->proxy_address_ = address;
		impl->proxy_port_ = port;
		impl->proxy_type_ = ptype;
		impl->proxy_login_ = login;
		impl->proxy_password_ = password;
		impl->connect_endpoint_cached_ = false;
	}

	void http_sync_connection::clear_proxy()
	{
		impl->proxy_type_ = proxy_none;
		impl->proxy_port_ = 0;
		impl->proxy_address_.clear();
		impl->proxy_login_.clear();
		impl->proxy_password_.clear();
		impl->connect_endpoint_cached_ = false;
	}

	void http_sync_connection::set_query_timeout(unsigned long timeout)
	{
		impl->query_timeout_ = timeout;
	}

	unsigned long http_sync_connection::get_query_timeout() const
	{
		return impl->query_timeout_;
	}

	void http_sync_connection::add_query_header(const std::string& name, const std::string& value)
	{
		impl->query_headers_.insert(std::make_pair(name, value));
	}

	bool http_sync_connection::remove_query_header(const std::string& name)
	{
		return impl->query_headers_.erase(name) > 0;
	}

	std::string http_sync_connection::get_query_header_value(const std::string& name) const
	{
		std::map<std::string, std::string>::const_iterator it = impl->query_headers_.find(name);
		if(it == impl->query_headers_.end())
			return "";

		return (*it).second;
	}

	void http_sync_connection::get_query_headers(std::map<std::string, std::string>& list) const
	{
		list.clear();
		list = impl->query_headers_;
	}

	void http_sync_connection::clear_query_headers()
	{
		impl->query_headers_.clear();
	}

	void http_sync_connection::set_auto_cookie_manager(cookie_manager& manager)
	{
		impl->cookie_mgr_ = &manager;
	}

	void http_sync_connection::remove_auto_cookie_manager()
	{
		impl->cookie_mgr_ = 0;
	}

	cookie_manager* http_sync_connection::get_auto_cookie_manager()
	{
		return impl->cookie_mgr_;
	}

	const cookie_manager* http_sync_connection::get_auto_cookie_manager() const
	{
		return impl->cookie_mgr_;
	}

	void http_sync_connection::enable_auto_redirect_follow(bool follow)
	{
		impl->follow_redirects_ = follow;

		impl->current_redirect_count_ = 0;
		impl->last_redirect_domain_.clear();
		impl->last_redirect_line_.clear();
		impl->last_redirect_port_ = 0;
		impl->last_redirect_uri_.clear();
	}

	bool http_sync_connection::auto_redirect_follow_enabled() const
	{
		return impl->follow_redirects_;
	}

	void http_sync_connection::enable_cross_domain_redirects(bool allow)
	{
		impl->allow_cross_domain_redirects_ = allow;
	}

	bool http_sync_connection::cross_diomain_redirects_enabled() const
	{
		return impl->allow_cross_domain_redirects_;
	}

	void http_sync_connection::set_max_redirect_count(unsigned long count)
	{
		impl->max_redirect_count_ = count;
	}

	unsigned long http_sync_connection::get_max_redirect_count() const
	{
		return impl->max_redirect_count_;
	}

	unsigned long http_sync_connection::get_recent_redirect_count() const
	{
		return impl->current_redirect_count_;
	}

	std::string http_sync_connection::get_last_redirect_line() const
	{
		return impl->last_redirect_line_;
	}

	std::string http_sync_connection::get_last_redirect_domain() const
	{
		return impl->last_redirect_domain_;
	}

	std::string http_sync_connection::get_last_redirect_uri() const
	{
		return impl->last_redirect_uri_;
	}

	unsigned short http_sync_connection::get_last_redirect_port() const
	{
		return impl->last_redirect_port_;
	}

	void http_sync_connection::enable_auto_urlencode(bool enable)
	{
		impl->auto_urlencode_ = enable;
	}

	bool http_sync_connection::auto_urlencode_enabled() const
	{
		return impl->auto_urlencode_;
	}


	std::string http_sync_connection::get(const http_endpoint& ep, const std::string& cookies, const std::string& referer)
	{
		set_connection_params(ep);
		return get(ep.uri(), ep.parameters(), cookies, referer);
	}

	std::string http_sync_connection::post(const http_endpoint& ep, const std::string& cookies, const std::string& referer)
	{
		set_connection_params(ep);
		return post(ep.uri(), ep.parameters(), cookies, referer);
	}

	std::string http_sync_connection::query(const std::string& method, const http_endpoint& ep, const std::string& cookies, const std::string& referer)
	{
		set_connection_params(ep);
		return query(ep.uri(), ep.parameters(), cookies, referer);
	}

	std::string http_sync_connection::multipart_post(const http_endpoint& ep, const std::string& multipart_file_parameter_name, multipart_file& file, const std::string& cookies, const std::string& referer)
	{
		set_connection_params(ep);
		return multipart_post(ep.uri(), multipart_file_parameter_name, file, ep.parameters(), cookies, referer);
	}

	std::string http_sync_connection::multipart_post(const http_endpoint& ep, std::map<std::string, multipart_file*>& files, const std::string& cookies, const std::string& referer)
	{
		set_connection_params(ep);
		return multipart_post(ep.uri(), files, ep.parameters(), cookies, referer);
	}

	std::string http_sync_connection::multipart_post(const http_endpoint& ep, const std::string& cookies, const std::string& referer)
	{
		set_connection_params(ep);
		return multipart_post(ep.uri(), ep.parameters(), cookies, referer);
	}

	void http_sync_connection::clear_http_load_callback()
	{
		load_callback_.swap(std::function<bool(const std::string&)>());
	}

	void http_sync_connection::clear_input_after_callback(bool clear)
	{
		impl->clear_input_after_callback_ = clear;
	}

	bool http_sync_connection::clear_input_after_callback() const
	{
		return impl->clear_input_after_callback_;
	}

	void http_sync_connection::set_callback_option(http_callback_option option)
	{
		impl->callback_option_ = option;
	}

	http_callback_option http_sync_connection::get_callback_option() const
	{
		return impl->callback_option_;
	}

	void http_sync_connection::set_callback_min_transfer_amount(unsigned long bytes)
	{
		impl->min_transfer_amount_ = bytes;
	}

	unsigned long http_sync_connection::get_callback_min_transfer_amount() const
	{
		return impl->min_transfer_amount_;
	}
}
