#pragma once
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include "simple_http_defs.h"
#include "simple_http_socket_worker_base.h"
#include "simple_http_impl.h"
#include "static_data.h"

namespace simple_http
{
	template<typename SocketObject>
	class simple_http_socket_worker : public simple_http_socket_worker_base
	{
	public:
		simple_http_socket_worker(boost::asio::io_service& ios, simple_http_impl* impl, std::function<bool(const std::string&)>& callback)
			:simple_http_socket_worker_base(ios, impl, callback), socket_(ios)
		{}

		virtual void timeout_handler(const boost::system::error_code& ec)
		{
			if(!ec)
				socket_.base_socket().close();
		}

		virtual void connect(boost::asio::ip::tcp::resolver::iterator endpoint_iterator)
		{
			if(impl_->connect_cache_enabled_ && !impl_->connect_endpoint_cached_)
			{
				if(impl_->proxy_type_ == proxy_socks5)
					boost::asio::async_connect(socket_.connect_socket(), endpoint_iterator, boost::bind(&simple_http_socket_worker::handle_connect_socks5_pre_auth, this, boost::asio::placeholders::error, boost::asio::placeholders::iterator));
				else if(impl_->proxy_type_ == proxy_http)
					boost::asio::async_connect(socket_.connect_socket(), endpoint_iterator, boost::bind(&simple_http_socket_worker::handle_connect_http_proxy_pre_auth, this, boost::asio::placeholders::error, boost::asio::placeholders::iterator));
				else
					boost::asio::async_connect(socket_.connect_socket(), endpoint_iterator, boost::bind(&simple_http_socket_worker::handle_connect, this, boost::asio::placeholders::error, boost::asio::placeholders::iterator));
			}
			else
			{
				if(impl_->proxy_type_ == proxy_socks5)
					socket_.async_connect(impl_->cached_endpoint_, boost::bind(&simple_http_socket_worker::handle_connect_socks5_pre_auth, this, boost::asio::placeholders::error, boost::asio::ip::tcp::resolver::iterator()));
				else if(impl_->proxy_type_ == proxy_http)
					socket_.async_connect(impl_->cached_endpoint_, boost::bind(&simple_http_socket_worker::handle_connect_http_proxy_pre_auth, this, boost::asio::placeholders::error, boost::asio::ip::tcp::resolver::iterator()));
				else
					socket_.async_connect(impl_->cached_endpoint_, boost::bind(&simple_http_socket_worker::handle_connect, this, boost::asio::placeholders::error, boost::asio::ip::tcp::resolver::iterator()));
			}
		}

		virtual void read_socks_answer(std::size_t transfer, socks5_auth_step next_step)
		{
			boost::asio::async_read(socket_.base_socket(), boost::asio::buffer(socks5_buf, sizeof(socks5_buf)), boost::asio::transfer_exactly(transfer), boost::bind(&simple_http_socket_worker::handle_connect_socks5_auth, this, boost::asio::placeholders::error, next_step));
		}

		virtual void write_socks_answer(socks5_auth_step next_step)
		{
			boost::asio::async_write(socket_.base_socket(), request_, boost::bind(&simple_http_socket_worker::handle_connect_socks5_auth, this, boost::asio::placeholders::error, next_step));
		}

		virtual void handle_connect(const boost::system::error_code& ec, boost::asio::ip::tcp::resolver::iterator endpoint_iterator)
		{
			if(ec)
			{
				check_for_timeout(ec);
				impl_->throw_connection_error();
			}

			cache_endpoint(endpoint_iterator);
			socket_.async_handshake(boost::bind(&simple_http_socket_worker::handle_handshake, this, boost::asio::placeholders::error));
		}

		virtual void write_request()
		{
			boost::asio::async_write(socket_, request_, boost::asio::transfer_all(), boost::bind(&simple_http_socket_worker::handle_write, this, boost::asio::placeholders::error));
		}

		virtual void write_file_data(bool write_next)
		{
			boost::asio::async_write(socket_, request_, boost::asio::transfer_all(), boost::bind(&simple_http_socket_worker::handle_write_files, this, boost::asio::placeholders::error, write_next));
		}

		virtual void handle_write(const boost::system::error_code& ec)
		{
			if(ec)
			{
				check_for_timeout(ec);
				throw http_error("Error writing to socket", http_error::error_writing_to_socket);
			}

			if(load_callback_)
			{
				if(impl_->callback_option_ == get_everything)
				{
					boost::asio::async_read(socket_, *out_, boost::asio::transfer_at_least(impl_->min_transfer_amount_), boost::bind(&simple_http_socket_worker::handle_read, this, boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred));
				}
				else
				{
					if(first_read_)
						boost::asio::async_read_until(socket_, *out_, headers_end_regex, boost::bind(&simple_http_socket_worker::handle_read, this, boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred));
					else
						boost::asio::async_read(socket_, *out_, boost::asio::transfer_at_least(impl_->min_transfer_amount_), boost::bind(&simple_http_socket_worker::handle_read, this, boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred));
				}
			}
			else
			{
				//TODO: parse content-length
				boost::asio::async_read(socket_, *out_, boost::asio::transfer_all(), boost::bind(&simple_http_socket_worker::handle_read, this, boost::asio::placeholders::error, boost::asio::placeholders::bytes_transferred));
			}
		}
		
		virtual void read_http_proxy_answer(https_proxy_auth_step next_step)
		{
			clean_request_buffer();
			boost::asio::async_read_until(socket_.base_socket(), request_, headers_end_regex, boost::bind(&simple_http_socket_worker::handle_connect_http_proxy_auth, this, boost::asio::placeholders::error, next_step));
		}

		virtual void write_http_proxy_answer(https_proxy_auth_step next_step)
		{
			boost::asio::async_write(socket_.base_socket(), request_, boost::bind(&simple_http_socket_worker::handle_connect_http_proxy_auth, this, boost::asio::placeholders::error, next_step));
		}

	private:
		SocketObject socket_;
	};
}
