#include <string>
#include <map>
#include <boost/asio.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/bind.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/random.hpp>
#include <functional>
#ifdef SSL_SUPPORT
#include <boost/asio/ssl.hpp>
#include "functions.h"
#endif
#include "simple_http_socket_worker_base.h"
#include "http_error.h"
#include "simple_http_impl.h"
#include "simple_http_defs.h"
#include "multipart_file.h"
#include "utils.h"
#include "cookie_manager.h"

namespace simple_http
{
	simple_http_socket_worker_base::simple_http_socket_worker_base(boost::asio::io_service& ios, simple_http_impl* impl, std::function<bool(const std::string&)>& callback)
		:impl_(impl), timer_(ios), load_callback_(callback), read_ptr_(0), first_read_(true)
	{}

	simple_http_socket_worker_base::~simple_http_socket_worker_base()
	{
		clean_request_buffer();
	}

	void simple_http_socket_worker_base::check_for_timeout(const boost::system::error_code& ec)
	{
		if(ec == boost::asio::error::operation_aborted)
			throw http_error("Timeout", http_error::timeout);
	}

	void simple_http_socket_worker_base::query(const std::string& query_type, boost::asio::streambuf& out, const std::string& uri, const std::string& parameters, const std::string& cookies, const std::string& referer, bool multipart, std::map<std::string, multipart_file*>* files)
	{
		out_ = &out;
		query_type_ = &query_type;
		uri_ = &uri;
		parameters_ = &parameters;
		cookies_ = &cookies;
		referer_ = &referer;
		multipart_ = multipart;
		files_ = files;

		timer_.expires_from_now(boost::posix_time::seconds(impl_->query_timeout_));
		timer_.async_wait(boost::bind(&simple_http_socket_worker_base::timeout_handler, this, boost::asio::placeholders::error));

		if(!impl_->connect_endpoint_cached_ || !impl_->connect_cache_enabled_)
		{
			boost::asio::ip::tcp::resolver::query query(boost::asio::ip::tcp::v4(),
				(impl_->proxy_type_ != proxy_none) ? impl_->proxy_address_ : impl_->address_,
				boost::lexical_cast<std::string>((impl_->proxy_type_ != proxy_none) ? impl_->proxy_port_ : impl_->port_));

			resolver_.reset(new boost::asio::ip::tcp::resolver(timer_.get_io_service()));
			resolver_->async_resolve(query, boost::bind(&simple_http_socket_worker_base::handle_resolve, this, boost::asio::placeholders::error, boost::asio::placeholders::iterator));
		}
		else
		{
			connect(boost::asio::ip::tcp::resolver::iterator());
		}
	}

	void simple_http_socket_worker_base::clean_request_buffer()
	{
		request_.consume(request_.size());
		request_.prepare(128);
	}

	void simple_http_socket_worker_base::cache_endpoint(boost::asio::ip::tcp::resolver::iterator endpoint_iterator)
	{
		if(impl_->connect_cache_enabled_ && endpoint_iterator != boost::asio::ip::tcp::resolver::iterator())
		{
			impl_->connect_endpoint_cached_ = true;
			impl_->cached_endpoint_ = *endpoint_iterator;
		}
	}

	void simple_http_socket_worker_base::handle_resolve(const boost::system::error_code& ec, boost::asio::ip::tcp::resolver::iterator endpoint_iterator)
	{
		if(ec || endpoint_iterator == boost::asio::ip::tcp::resolver::iterator())
		{
			check_for_timeout(ec);
			throw http_error("Address resolving error", http_error::cannot_resolve);
		}

		connect(endpoint_iterator);
	}

	void simple_http_socket_worker_base::handle_connect_socks5_pre_auth(const boost::system::error_code& ec, boost::asio::ip::tcp::resolver::iterator endpoint_iterator)
	{
		if(ec)
		{
			check_for_timeout(ec);
			impl_->throw_connection_error();
		}

		cache_endpoint(endpoint_iterator);

		handle_connect_socks5_auth(boost::system::error_code(), socks5_write_auth_types_request);
	}

	void simple_http_socket_worker_base::handle_connect_http_proxy_pre_auth(const boost::system::error_code& ec, boost::asio::ip::tcp::resolver::iterator endpoint_iterator)
	{
		if(ec)
		{
			check_for_timeout(ec);
			impl_->throw_connection_error();
		}

		cache_endpoint(endpoint_iterator);

		handle_connect_http_proxy_auth(boost::system::error_code(), https_proxy_send_connect);
	}

	void simple_http_socket_worker_base::handle_connect_http_proxy_auth(const boost::system::error_code& ec, https_proxy_auth_step step)
	{
		if(ec && ec != boost::asio::error::eof)
		{
			check_for_timeout(ec);
			impl_->throw_connection_error();
		}

#ifdef SSL_SUPPORT
		if(impl_->ssl_)
		{
			if(step != 2)
				clean_request_buffer();

			std::ostream stream(&request_);

			switch(step)
			{
			case https_proxy_send_connect:
				stream << "CONNECT " << impl_->address_ << ":" << impl_->port_ << " HTTP/1.0\r\n";
				if(!impl_->proxy_login_.empty() || !impl_->proxy_password_.empty())
					stream << "Proxy-Authorization: Basic " << utils::text_to_base64(impl_->proxy_login_ + ":" + impl_->proxy_password_) << "\r\n";

				stream << "\r\n";

				write_http_proxy_answer(https_connect_read_connect_answer);
				return;

			case https_connect_read_connect_answer:
				read_http_proxy_answer(https_proxy_check_connect_answer);
				return;

			case https_proxy_check_connect_answer:
				{
					const char* header = boost::asio::buffer_cast<const char*>(request_.data());
					std::string proxy_result(header, request_.size());
					clean_request_buffer();
					unsigned long code = get_answer_code(proxy_result);

					if(boost::icontains(proxy_result, "Connection established") && code == 200)
					{
						handle_connect(boost::system::error_code(), boost::asio::ip::tcp::resolver::iterator());
					}
					else
					{
						if(code == 407)
							throw http_error("HTTPS proxy authorization error", http_error::cannot_authorize_to_https_proxy);
						else
							throw http_error("HTTPS proxy connection error", http_error::cannot_connect_to_https_proxy);
					}
				}
				return;
			}
		}
		else
#endif
		{
			handle_connect(boost::system::error_code(), boost::asio::ip::tcp::resolver::iterator());
		}
	}

	void simple_http_socket_worker_base::handle_read(const boost::system::error_code& ec, std::size_t bytes)
	{
		if(ec && ec != boost::asio::error::eof)
		{
			check_for_timeout(ec);

#ifdef SSL_SUPPORT
			if(ec.value() == static_cast<int>(ERR_PACK(ERR_LIB_SSL, 0, SSL_R_SHORT_READ)))
			{
				if(bytes)
				{
					handle_write(boost::system::error_code());
					return;
				}
			}
			else
#endif
			{
				throw http_error("Error reading from socket", http_error::error_reading_from_socket);
			}
		}
		else if(load_callback_ && bytes)
		{
			bool next = true;

			{
				const char* header = boost::asio::buffer_cast<const char*>(out_->data());

				if(impl_->callback_option_ == get_everything)
				{
					next = load_callback_(std::string(header + read_ptr_, bytes));
				}
				else if(impl_->callback_option_ == get_headers_only && first_read_)
				{
					next = load_callback_(std::string(header, bytes));
				}
				else if(impl_->callback_option_ == get_body_only)
				{
					if(out_->size() > bytes || !first_read_)
					{
						std::string ret = first_read_ ? 
							std::string(header + bytes, out_->size() - bytes) :
						std::string(header + read_ptr_, bytes);

						if(first_read_)
							read_ptr_ = out_->size() - bytes;

						next = load_callback_(ret);
					}
				}
			}


			read_ptr_ += bytes;

			if(impl_->clear_input_after_callback_)
			{
				out_->consume(out_->size());
				if(impl_->callback_option_ == get_everything || impl_->callback_option_ == get_body_only)
					read_ptr_ = 0;
				else if(impl_->callback_option_ == get_headers_only)
					next = false;
			}

			first_read_ = false;


			if(next)
			{
				handle_write(boost::system::error_code());
				return;
			}
		}
		
		timer_.cancel(boost::system::error_code(boost::asio::error::operation_aborted));
	}

	void simple_http_socket_worker_base::handle_connect_socks5_auth(const boost::system::error_code& ec, socks5_auth_step step)
	{
		if(ec)
		{
			check_for_timeout(ec);
			throw http_error("Error connecting to socks5 server", http_error::cannot_connect_to_socks5_proxy);
		}

		clean_request_buffer();
		std::ostream stream(&request_);

		switch(step)
		{
		case socks5_write_auth_types_request:
			stream.write("\x05\x02\x00\x02", 4);
			write_socks_answer(socks5_read_auth_types_request);
			return;

		case socks5_read_auth_types_request:
			read_socks_answer(2, socks5_auth);
			return;

		case socks5_auth:
			if(*reinterpret_cast<unsigned short*>(&socks5_buf[0]) == 0x205)
			{
				if(impl_->proxy_login_.empty() || impl_->proxy_password_.empty())
					throw http_error("Socks5 server requires login/password", http_error::socks5_auth_required);

				if(impl_->proxy_login_.length() > 0xFF || impl_->proxy_password_.length() > 0xFF)
					throw http_error("Too long login/password for socks5", http_error::socks5_too_long_login_or_password);

				//Authorize
				stream.write("\x01", 1);
				unsigned char length = static_cast<unsigned char>(impl_->proxy_login_.length());
				stream.write(reinterpret_cast<const char*>(&length), sizeof(unsigned char));
				stream << impl_->proxy_login_;
				length = static_cast<unsigned char>(impl_->proxy_password_.length());
				stream.write(reinterpret_cast<const char*>(&length), sizeof(unsigned char));
				stream << impl_->proxy_password_;

				write_socks_answer(socks5_read_auth_answer);
				return;
			}
			else if(*reinterpret_cast<unsigned short*>(&socks5_buf[0]) != 5)
			{
				throw http_error("Can't connect to socks5 server", http_error::cannot_connect_to_socks5_proxy);
			}

			handle_connect_socks5_auth(boost::system::error_code(), socks5_connect_request);
			return;

		case socks5_read_auth_answer:
			read_socks_answer(2, socks5_check_auth_answer);
			return;

		case socks5_check_auth_answer:
			if(socks5_buf[1] != 0)
				throw http_error("Can't authorize with socks5 server", http_error::socks5_authorization_error);

			handle_connect_socks5_auth(boost::system::error_code(), socks5_connect_request);
			break;

		case socks5_connect_request:
			{
				unsigned char length = static_cast<unsigned char>(impl_->address_.length());
				stream.write("\x05\x01\x00\x03", 4);
				stream.write(reinterpret_cast<const char*>(&length), sizeof(unsigned char));
				stream << impl_->address_;
				stream.write(reinterpret_cast<const char*>(&impl_->port_) + 1, sizeof(unsigned char));
				stream.write(reinterpret_cast<const char*>(&impl_->port_), sizeof(unsigned char));

				write_socks_answer(socks5_read_connect_answer);
			}
			return;

		case socks5_read_connect_answer:
			read_socks_answer(10, socks5_check_connect_answer);
			return;

		case socks5_check_connect_answer:
			if(socks5_buf[4] == 0)
				throw http_error("Can't connect to site via socks5 server", http_error::socks5_site_connection_error);

			handle_connect(boost::system::error_code(), boost::asio::ip::tcp::resolver::iterator());
			return;
		}
	}

	void simple_http_socket_worker_base::handle_handshake(const boost::system::error_code& ec)
	{
		if(ec)
		{
			check_for_timeout(ec);
			throw http_error("SSL Handshake error", http_error::ssl_handshake_error);
		}

		clean_request_buffer();
		std::ostream stream(&request_);

		stream << *query_type_ << " ";

		if(!impl_->ssl_ && impl_->proxy_type_ == proxy_http)
			stream << "http://" << impl_->address_ << ":" << impl_->port_;

		stream << *uri_;

		if(!parameters_->empty() && *query_type_ != "POST")
			stream << "?" << (impl_->auto_urlencode_ ? utils::urlencode_param_string_ansi(*parameters_) : *parameters_);

		stream << " HTTP/1.0\r\n";

		if(!impl_->ssl_ && impl_->proxy_type_ == proxy_http && (!impl_->proxy_login_.empty() || !impl_->proxy_password_.empty()))
			stream << "Proxy-Authorization: Basic " << utils::text_to_base64(impl_->proxy_login_ + ":" + impl_->proxy_password_) << "\r\n";

		stream << "Host: " << impl_->address_;

		if((impl_->port_ != 80 && !impl_->ssl_) || (impl_->port_ != 443 && impl_->ssl_))
			stream << ":" << impl_->port_;

		stream << "\r\n";

		if(!impl_->user_agent_.empty())
			stream << "User-agent: " << impl_->user_agent_ << "\r\n";

		{
			std::string cookie_line;

			if(impl_->cookie_mgr_)
			{
				impl_->cookie_mgr_->remove_expired_cookies();
				cookie_line.append(impl_->cookie_mgr_->get_cookies(impl_->address_, utils::get_path_from_uri(*uri_), impl_->ssl_ ? connection_secure : connection_http));
			}

			cookie_line.append(*cookies_);

			if(!cookie_line.empty())
				stream << "Cookie: " << cookie_line << "\r\n";
		}

		if(!referer_->empty())
			stream << "Referer: " << *referer_ << "\r\n";

		if(!impl_->query_headers_.empty())
		{
			for(std::map<std::string, std::string>::const_iterator it = impl_->query_headers_.begin(); it != impl_->query_headers_.end(); ++it)
				stream << (*it).first << ": " << (*it).second << "\r\n";
		}

		stream << "Connection: close\r\n";


		if(*query_type_ == "POST")
		{
			if(!multipart_)
			{
				if(impl_->auto_urlencode_)
				{
					std::string params_urlencoded(utils::urlencode_param_string_ansi(*parameters_));

					stream << "Content-Type: application/x-www-form-urlencoded\r\n"
						<< "Content-Length: " << params_urlencoded.length() << "\r\n\r\n";

					stream << params_urlencoded;
				}
				else
				{
					stream << "Content-Type: application/x-www-form-urlencoded\r\n"
						<< "Content-Length: " << parameters_->length() << "\r\n\r\n";

					stream << *parameters_;
				}
			}
			else
			{
				boundary_ = "-----------------";

				{
					boost::random::mt19937 rng;
					unsigned long generated = 0;
					rng.generate(&generated, &generated + 1);
					boundary_.append(boost::lexical_cast<std::string>(generated));
				}

				stream << "Content-Type: multipart/form-data; boundary=" << boundary_ << "\r\n";

				std::vector<std::string> result;
				if(!parameters_->empty())
					boost::split(result, *parameters_, boost::algorithm::is_from_range('&', '&'), boost::algorithm::token_compress_on);

				{
					unsigned long total_parameter_length = 0;
					for(std::vector<std::string>::const_iterator it = result.begin(); it != result.end(); ++it)
					{
						total_parameter_length += (*it).length();

						if((*it).find('=') != std::string::npos)
							total_parameter_length--;
					}


					if(files_)
					{
						for(std::map<std::string, multipart_file*>::const_iterator it = files_->begin(); it != files_->end(); ++it)
						{
							total_parameter_length += (*it).first.length();
							total_parameter_length += (*it).second->get_name().length();
							total_parameter_length += (*it).second->get_length();
						}
					}

					unsigned long multipart_length =
						(result.size()) *
						(sizeof("Content-Disposition: form-data; name=\"\"\r\n\r\n") - 1/*nullbyte*/ + boundary_.length() + 2 /*--*/ + 2 /*\r\n after boundary*/ + 2 /*\r\n after data*/)
						+
						(files_ ? files_->size() : 0) *
						(sizeof("Content-Disposition: form-data; name=\"\"; filename=\"\"\r\n\r\n") - 1/*nullbyte*/ + boundary_.length() + 2 /*--*/ + 2 /*\r\n afrer bounday*/ + 2 /*\r\n after filedata*/)
						+
						boundary_.length() + 4 /*--boundary--*/
						+
						total_parameter_length
						+
						2 /*\r\n ending*/;

					stream << "Content-Length: " << multipart_length << "\r\n\r\n";
				}

				for(std::vector<std::string>::const_iterator it = result.begin(); it != result.end(); ++it)
				{
					std::string::size_type pos = (*it).find('=');

					stream << "--" << boundary_ << "\r\n";

					if(pos != std::string::npos)
					{
						stream << "Content-Disposition: form-data; name=\"" << (*it).substr(0, pos) << "\"\r\n\r\n"
							<< (*it).substr(pos + 1) << "\r\n";
					}
					else
					{
						stream << "Content-Disposition: form-data; name=\"" << (*it) << "\"\r\n\r\n\r\n"; //empty value
					}
				}

				if(files_ && files_->size() > 0)
				{
					file_iterator_ = files_->begin();
					write_file_data(true);
					return;
				}

				stream << "--" << boundary_ << "--\r\n";
			}
		}
		else
		{
			stream << "\r\n";
		}

		write_request();
	}


	void simple_http_socket_worker_base::handle_write_files(const boost::system::error_code& ec, bool write_next)
	{
		if(ec)
		{
			check_for_timeout(ec);
			throw http_error("Error writing to socket", http_error::error_writing_to_socket);
		}

		clean_request_buffer();
		std::ostream stream(&request_);

		if(write_next)
		{
			if(file_iterator_ == files_->end())
			{
				stream << "--" << boundary_ << "--\r\n";
				write_request();
				return;
			}

			stream << "--" << boundary_ << "\r\n"
				<< "Content-Disposition: form-data; name=\"" << (*file_iterator_).first << "\"; filename=\""
				<< (*file_iterator_).second->get_name() << "\"\r\n\r\n";

			(*file_iterator_).second->reset_pointer();
		}

		{
			char buf[file_buffer_size];

			unsigned long read = (*file_iterator_).second->read_data(buf, sizeof(buf));

			if(read)
			{
				stream.write(buf, read);
				write_file_data(false);
				return;
			}
		}

		(*file_iterator_).second->close();

		++file_iterator_;
		stream << "\r\n";
		write_file_data(true);
	}
}
