#include <string>
#include <boost/regex.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/foreach.hpp>
#include <boost/date_time/gregorian/gregorian.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/algorithm/string.hpp>
#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#endif
#include "utils.h"
#include "static_data.h"
#include "base64.h"

namespace simple_http
{
	namespace utils
	{
		tm cookie_date_to_tm(const std::string& cookie_date)
		{
			//Tue, 15-Jan-2013 21:47:38 GMT
			static const std::string months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
			boost::match_results<std::string::const_iterator> what;

			try
			{
				if(boost::regex_match(cookie_date, what, cookie_expiry_date))
				{
					try
					{
						unsigned short day = boost::lexical_cast<unsigned short>(what[1]);
						unsigned short year = boost::lexical_cast<unsigned short>(what[3]);
						unsigned short hour = boost::lexical_cast<unsigned short>(what[4]);
						unsigned short minute = boost::lexical_cast<unsigned short>(what[5]);
						unsigned short second = boost::lexical_cast<unsigned short>(what[6]);
						unsigned short month = 0;
						std::string month_str = what[2].str();
						bool month_found = false;
						BOOST_FOREACH(const std::string& s, months)
						{
							month++;
							if(s == month_str)
							{
								month_found = true;
								break;
							}
						}

						if(!month_found)
							return tm();

						if(year < 50)
							year += 2000;
						if(year < 100)
							year += 1900;

						return boost::posix_time::to_tm(
							boost::posix_time::ptime(boost::gregorian::date(year, month, day), boost::posix_time::time_duration(hour, minute, second))
							);
					}
					catch(const boost::bad_lexical_cast&)
					{
						return tm();
					}
					catch(const boost::gregorian::bad_month&)
					{
						return tm();
					}
					catch(const boost::gregorian::bad_year&)
					{
						return tm();
					}
					catch(const boost::gregorian::bad_day_of_month&)
					{
						return tm();
					}
				}
			}
			catch(const boost::regex_error&)
			{
				return tm();
			}

			return tm();
		}

		std::string text_to_base64(const std::string& text)
		{
			if(text.empty())
				return "";

			return base64::base64_encode(reinterpret_cast<const unsigned char*>(text.c_str()), text.length());
		}

		std::string data_to_base64(const char* data, unsigned long length)
		{
			if(!length)
				return "";

			return base64::base64_encode(reinterpret_cast<const unsigned char*>(data), length);
		}

		template<typename T>
		std::string urlencode(const T& s);

		template<>
		std::string urlencode<std::string>(const std::string& s)
		{
			std::stringstream ret;
			ret << std::hex << std::uppercase;
			for(std::string::const_iterator it = s.begin(); it != s.end(); ++it)
			{
				unsigned char c = *it;

				if(c < 0x80)
				{
					if(c == ' ')
						ret << "+";
					else if((c <= 47 || (c >= 58 && c <= 64) || (c >= 91 && c <= 96) || c >=123) && c != '.' && c != '_' && c != '-' && c != '~')
						ret << "%" << std::setw(2) << std::setfill('0') << static_cast<unsigned long>(c);
					else
						ret << c;
				}
				else
				{
					ret << "%" << std::setw(2) << std::setfill('0') << static_cast<unsigned long>(c);
				}
			}

			return ret.str();
		}

		template<>
		std::string urlencode<std::wstring>(const std::wstring& s)
		{
#ifdef WIN32
			if(s.empty())
				return "";

			int ret = WideCharToMultiByte(CP_UTF8, 0, s.c_str(), -1, 0, 0, 0, 0);
			if(!ret)
				return "";

			boost::scoped_array<char> buf(new char[ret + 1]);
			if(!WideCharToMultiByte(CP_UTF8, 0, s.c_str(), -1, buf.get(), ret, 0, 0))
				return "";

			return urlencode<std::string>(buf.get());
#else
			throw std::exception("Not supported in non-Windows builds");
#endif
		}

		std::string urlencode_unicode(const std::wstring& s)
		{
			return urlencode(s);
		}

		std::string urlencode_utf8(const std::string& s)
		{
			return urlencode(s);
		}

		std::string urlencode_ansi(const std::string& s)
		{
			return urlencode(s);
		}

		std::string base64_to_data(const std::string& base64)
		{
			return base64::base64_decode(base64);
		}


		std::string get_path_from_uri(const std::string& uri)
		{
			std::string ret("/");
			if(uri.empty())
				return ret;

			if(uri[0] == '/')
				ret.clear();

			std::string::size_type pos = uri.find_last_of('/');
			if(pos == std::string::npos)
				ret.append(uri);
			else
				ret.append(uri.substr(0, pos + 1));

			if(ret.back() != '/')
				ret.push_back('/');

			return ret;
		}


		template<typename T>
		std::string urlencode_param_string(const T& parameters)
		{
			std::string ret;
			if(parameters.empty())
				return ret;

			typename std::vector<T> result;
			boost::split(result, parameters, boost::algorithm::is_from_range('&', '&'), boost::algorithm::token_compress_on);

			for(typename std::vector<T>::const_iterator it = result.begin(); it != result.end(); ++it)
			{
				typename T::size_type pos = (*it).find('=');
				if(pos == T::npos)
					ret.append(urlencode<T>(*it)).push_back('&');
				else
					ret.append(urlencode<T>(T((*it).c_str(), pos))).append("=").append(urlencode<T>((*it).c_str() + pos + 1)).push_back('&');
			}

			return ret;
		}

		std::string urlencode_param_string_unicode(const std::wstring& parameters)
		{
			return urlencode_param_string(parameters);
		}

		std::string urlencode_param_string_utf8(const std::string& parameters)
		{
			return urlencode_param_string(parameters);
		}

		std::string urlencode_param_string_ansi(const std::string& parameters)
		{
			return urlencode_param_string(parameters);
		}

		bool is_ip_address(const std::string& domain)
		{
			try
			{
				boost::smatch what;
				if(!boost::regex_match(domain, what, ip_address_regex))
					return false;

				unsigned short a = boost::lexical_cast<unsigned short>(what[1].str()),
					b = boost::lexical_cast<unsigned short>(what[2].str()),
					c = boost::lexical_cast<unsigned short>(what[3].str()),
					d = boost::lexical_cast<unsigned short>(what[4].str());

				if(a > 255 || b > 255 || c > 255 || d > 255)
					return false;

				return true;
			}
			catch(const boost::regex_error&)
			{
				return false;
			}
			catch(const boost::bad_lexical_cast&)
			{
				return false;
			}
		}
	}
}
